/*
 *  filegrabber.cpp
 *
 *  Author: Domenico Bloisi
 *  domenicobloisi@yahoo.com
 *  Copyright 2010
 */

#include "filegrabber.h"

FileGrabber::FileGrabber(const char *fn, int prev_offset)
{
	prev_counter = 0;
	this->prev_offset = prev_offset;

	capture = cvCaptureFromFile(fn);
	if(!capture) {
		cout<<"unable to capture from "<<fn<<endl;
		cout<<"aborting..."<<endl;
		abort();
	}

	cvQueryFrame(capture);
	height = (int) cvGetCaptureProperty(capture, CV_CAP_PROP_FRAME_HEIGHT);
	width = (int) cvGetCaptureProperty(capture, CV_CAP_PROP_FRAME_WIDTH);
	fps = cvGetCaptureProperty(capture, CV_CAP_PROP_FPS);
	CvSize frameSize = cvSize(width, height);

	time_scale = 1000 / fps;

	frameNumber = 0;

	frame = cvCreateImage(frameSize, IPL_DEPTH_8U, 3);
	prevFrame = cvCreateImage(frameSize, IPL_DEPTH_8U, 3);

	timestamp = 0;
}

FileGrabber::~FileGrabber()
{
	cvReleaseImage(&frame);
	cvReleaseImage(&prevFrame);
}

bool FileGrabber::getData()
{
	prev_counter++;
	if(prev_counter == prev_offset) {
		cvCopy(frame, prevFrame);
		prev_counter = 0;
	}

	if(!cvGrabFrame( capture )) {
			return false;
	}

	cvCopy(cvRetrieveFrame(capture), frame);

	prevTimestamp = timestamp;
	timestamp = (long)(frameNumber * time_scale);

	frameNumber++;
	return true;
}

unsigned long FileGrabber::getTimestamp() {
	return timestamp;
}

unsigned long FileGrabber::getPrevTimestamp() {
	return prevTimestamp;
}
