/*
 *  imagegrabber.cpp
 *
 *  Author: Domenico Bloisi
 *  domenicobloisi@yahoo.com
 *  Copyright 2010
 */

#include "imagegrabber.h"
#include <sstream>

ImageGrabber::ImageGrabber(string fn, int d, int fps, int prev_offset, bool resize, CvSize dim) {
	this->d = d;
	this->fps = fps;
	this->prev_offset = prev_offset;
	this->resize = resize;
	prev_counter = 0;
	time_scale = 1000/this->fps;

	int index = fn.find_last_of(".");
	string s = fn.substr(index - d, d);

	int zeros = 0;
	for(int i = 0; i < (d - 1); i++) {
		if(s.substr(i, 1).compare("0") == 0) {
			zeros++;
		}
		else {
			break;
		}
	}

	img_path.assign(fn);

	IplImage* img = cvLoadImage(img_path.c_str());
	if( !img ) {
		cerr<<"unable to read image "<<img_path<<endl;
		cerr<<"aborting..."<<endl;
		abort();
	}

	if(resize) {
		width = dim.width;
		height = dim.height;
	}
	else {
		width = img->width;
		height = img->height;
	}
	cvReleaseImage(&img);

	CvSize frameSize = cvSize(width, height);

	istringstream buffer(s.substr(zeros));
	buffer >> frameNumber;

	frame = cvCreateImage(frameSize, IPL_DEPTH_8U, 3);
	prevFrame = cvCreateImage(frameSize, IPL_DEPTH_8U, 3);

	next_img_path.assign(img_path);

#if __linux__
	index = img_path.find_last_of("/");
#else
	index = img_path.find_last_of("\\");
#endif
	img_path.erase(0, index);

	timestamp = 0;
}

ImageGrabber::~ImageGrabber() {
	cvReleaseImage(&frame);
	cvReleaseImage(&prevFrame);
}

bool ImageGrabber::getData() {
	prev_counter++;
	if(prev_counter == prev_offset) {
		cvCopy(frame, prevFrame);
		prev_counter = 0;
	}

	IplImage* loadedImage = cvLoadImage(next_img_path.c_str());
	if(loadedImage == NULL) {
		return false;
	}
	if(resize) {
		cvResize(loadedImage, frame );
	}
	else {
		cvCopy(loadedImage, frame);
	}
	cvReleaseImage(&loadedImage);

	img_path.assign(next_img_path);

	int index = next_img_path.find_last_of(".");
	istringstream buffer(next_img_path.substr(index-d, d));
	int frame_number;
	buffer >> frame_number;

	prevTimestamp = timestamp;

	timestamp = (long)(frame_number * time_scale);

	frameNumber = frame_number;

	frame_number++;

	ostringstream out_buffer;
	out_buffer << frame_number;
	index_string = out_buffer.str();
	int length = index_string.length();

	for(int i = d; i > length; i--) {
		index_string.insert(0, "0");
	}

	next_img_path.replace(index-d, d, index_string);

	return true;
}

string ImageGrabber::getImagePath() {
	return img_path;
}

unsigned long ImageGrabber::getTimestamp() {
	return timestamp;
}

unsigned long ImageGrabber::getPrevTimestamp() {
	return prevTimestamp;
}
