/*
 *  imbs.cpp
 *
 *  Author: Domenico Bloisi
 *  domenicobloisi@yahoo.com
 *  Copyright 2010
 */

#include "imbs.h"
#include "imagestereograbber.h"

IMBS::IMBS(Grabber *g,
		int fgThreshold,
		int binWidth,
		unsigned long samplingPeriod,
		int totalImageParts,
		int minBinHeight,
		int	numSamples,
		double alpha,
		double beta,
		double tau_s,
		double tau_h,
		double area,
		IMBS *stereoImbs)
{
	grabber = g;
	frameWidth = grabber->getWidth();
	frameHeight = grabber->getHeight();
	numPixels = frameHeight * frameWidth;
	CvSize frameSize = cvSize(frameWidth, frameHeight);
	frame = grabber->frame;

	hsvFrame = cvCreateImage( frameSize, IPL_DEPTH_32F, 3 );
	hsvMat = cvCreateImage(cvSize(1,1), IPL_DEPTH_32F, 3);
	rgbMat = cvCreateImage(cvSize(1,1), IPL_DEPTH_32F, 3);
	depthBlobs = cvCreateImage(cvSize(640, 480), 8, 3);

	fgImage = cvCreateImage( frameSize, 8, 1 );

	fgFilteredImage = cvCreateImage( frameSize, 8, 1 );

	fgColoredImage = cvCreateImage( frameSize, 8, 3 );

	bgValues = new BgStruct[numPixels];
    prevBgValues = new BgStruct[numPixels];
	bgModel = new BgModel[numPixels];
    //bgModelPrev = new BgModel[numPixels];

	if(minBinHeight <= 0){
		minBinHeight = 1;
	}

	maxBgValues = numSamples / minBinHeight;

	for(int i = 0; i < numPixels; i++){
		bgValues[i].currentValue = new Pixel[numSamples];
		bgValues[i].counter = new char[numSamples];
        prevBgValues[i].currentValue = new Pixel[numSamples];
        prevBgValues[i].counter = new char[numSamples];
		bgModel[i].values = new Pixel[maxBgValues];
        bgModel[i].isValid = new bool[maxBgValues];
        bgModel[i].isValid[0] = false;
        bgModel[i].counter = new int[maxBgValues];
	}

	bgSample = cvCreateImage( frameSize, IPL_DEPTH_8U, 3 );

	bgImage = cvCreateImage( frameSize, IPL_DEPTH_8U, 3 );

    bgStereoImageMat = cv::Mat::zeros(cv::Size(frameSize.width, frameSize.height), CV_16UC1);

	this->fgThreshold = fgThreshold;

	this->binWidth = binWidth;

	this->samplingPeriod = samplingPeriod;

	this->totalImageParts = totalImageParts;

	this->minBinHeight = minBinHeight;

	this->numSamples = numSamples;

	this->alpha = alpha;
	this->beta = beta;
	this->tau_s = tau_s;
	this->tau_h = tau_h;

	this->area = area;

	prev_bg_frame_time = 0;
	bg_frame_counter = 0;
	image_part_index = 0;

	bgFilename = NULL;

	bg_reset = false;

	prev_area = 0;

	stereo = false;

	this->stereoImbs = stereoImbs;

	isCreated = false;

}

IMBS::IMBS(Grabber *g,
		int fgThreshold,
		int binWidth,
		unsigned long samplingPeriod,
		int totalImageParts,
		int minBinHeight,
		int	numSamples) {

	grabber = g;
	frameWidth = grabber->getWidth();
	frameHeight = grabber->getHeight();
	numPixels = frameHeight * frameWidth;
	CvSize frameSize = cvSize(frameWidth, frameHeight);



    frameStereo = static_cast<imagestereograbber*>(grabber)->getDepth16bit();

    bgValues = new BgStruct[numPixels];
    bgModel = new BgModel[numPixels];

    disparity = new short[frameHeight*frameWidth*2];
	if(disparity == NULL) {
		cout << "unable to allocate disparity" << endl;
		exit(EXIT_FAILURE);
	}



	if(minBinHeight <= 0){
		minBinHeight = 1;
	}

	maxBgValues = numSamples / minBinHeight;

	for(int i = 0; i < numPixels; i++){
		bgValues[i].currentValue = new Pixel[numSamples];
		bgValues[i].counter = new char[numSamples];
		bgModel[i].values = new Pixel[maxBgValues];
		bgModel[i].isValid = new bool[maxBgValues];
		bgModel[i].isValid[0] = false;
		bgModel[i].counter = new int[maxBgValues];
	}

	bgSample = cvCreateImage( frameSize, IPL_DEPTH_8U, 3);

	bgImage = cvCreateImage( frameSize, IPL_DEPTH_8U, 3 );

    bgStereoImageMat = cv::Mat::zeros(cv::Size(frameSize.width, frameSize.height), CV_16UC1);

	this->fgThreshold = fgThreshold;

	this->binWidth = binWidth;

	this->samplingPeriod = samplingPeriod;

	this->totalImageParts = totalImageParts;

	this->minBinHeight = minBinHeight;

	this->numSamples = numSamples;

	prev_bg_frame_time = 0;
	bg_frame_counter = 0;
	image_part_index = 0;

	bgFilename = NULL;

	bg_reset = false;

	prev_area = 0;

	stereo = true;

	isCreated = false;

    bgStereoFilename = "";

}

IMBS::~IMBS() {
	cvReleaseImage(&frame);
	cvReleaseImage(&hsvFrame);
	cvReleaseImage(&fgImage);
	cvReleaseImage(&fgFilteredImage);
	cvReleaseImage(&bgSample);
	cvReleaseImage(&bgImage);
	cvReleaseImage(&hsvMat);
	cvReleaseImage(&rgbMat);
	cvReleaseImage(&grayForeground);
	cvReleaseImage(&whiteForeground);
    cvReleaseImage(&depthBlobs);

	delete[] disparity;
	delete[] bgValues;
	delete[] bgModel;
	delete[] fgStereo;
}

void IMBS::setSaveBackground(string* filename) {
	bgFilename = filename;
}

void IMBS::setSaveStereoBackground(string filename) {
	bgStereoFilename = filename;
}

void IMBS::getStereoForeground()  {
	bool isForeground;

    for(int c = 0; c < numPixels; c++) {
		isForeground = true;

		for(int n = 0; n < maxBgValues; n++) {
			if(!stereoImbs->bgModel[c].isValid[n]) {
				if(n == 0) {
					isForeground = false;
				}
				break;
			}
			else {
                int d = abs(stereoImbs->bgModel[c].values[n].d - currentDisparity[c]);
				if(d < fgThreshold){
					isForeground = false;
				}
			}
		}

		if(isForeground) {
			fgStereo[c] = 255;
		} else {
			fgStereo[c] = 0;
		}

	}
}


void IMBS::getForeground() {
    cvZero(fgImage);

    bool isForeground = true;

    for(int c = 0, k = 0; c < numPixels; k+=3, c++) {

        isForeground = true;

        for(int n = 0; n < maxBgValues; n++) {
            if(!bgModel[c].isValid[n]) {

                if(n == 0) {
                    isForeground = false;
                }

                break;
            }
            else {
                int d = max(abs(bgModel[c].values[n].b - (uchar)frame->imageData[k]),
                        abs(bgModel[c].values[n].g - (uchar)frame->imageData[k+1]) );
                d = max(d, abs(bgModel[c].values[n].r - (uchar)frame->imageData[k+2]) );
                if(d < fgThreshold){
                	isForeground = false;
                }
            }
        }
        if(isForeground) {
        	fgImage->imageData[c] = (uchar)255;
        }
    }
}

void IMBS::updateBackground() {

	if(bg_reset) {
		if(bg_frame_counter > numSamples - 1) {
			bg_frame_counter = numSamples - 1;
		}
	}

	if(prev_bg_frame_time > grabber->getTimestamp()) {
		prev_bg_frame_time = grabber->getTimestamp();
	}
	if(bg_frame_counter == numSamples - 1) {
		if(image_part_index < totalImageParts) {
			if(!stereo)
			createBackground(bg_frame_counter,
						image_part_index,
						true);
            else
				createStereoBackground(bg_frame_counter, image_part_index, true);

			image_part_index++;
		}
		else if(image_part_index == totalImageParts) {
			bg_frame_counter = 0;
			image_part_index = 0;
		}
	}
	else {

		if(((grabber->getTimestamp() - prev_bg_frame_time) >= samplingPeriod) &&
			(image_part_index == totalImageParts))
		{
			prev_bg_frame_time = grabber->getTimestamp();
			if(!stereo)
				cvCopy(frame, bgSample);
			else {
                for(int i = 0; i < frameHeight; ++i) {
                    for(int j = 0; j < frameWidth; ++j) {
                            disparity[i * frameWidth + j] =  static_cast<imagestereograbber*>(grabber)->getDepth16bit().at<ushort>(i,j);
                    }
                }

			}

		    image_part_index = 0;
			bg_frame_counter++;
		}
		else if(image_part_index < totalImageParts)	{
			if(!stereo)
				createBackground(bg_frame_counter, image_part_index, false);
            else {
				createStereoBackground(bg_frame_counter, image_part_index, false);
            }
			image_part_index++;
		}
	}
}

void IMBS::createStereoBackground(int frame_number, int image_interval, bool do_background) {
	if(bgSample == 0) {
		cerr << "createStereoBackground -- an error occurred: " <<
				" unable to retrieve frame no. " << frame_number << endl;
		abort();
	}

	int h_interval = frameHeight/totalImageParts;
	int h = (image_interval * h_interval);
    //int image_index = h * bgStereoImage->width;
    int image_index = h * bgStereoImageMat.cols;
	Pixel currentPixel;

    for(int t = 0; t < h_interval; h++, t++, image_index += bgStereoImageMat.cols) {
        for(int j = 0, image_cell = image_index, c = image_index;
				j < frameWidth;
            j++, image_cell++, c++)
		{
//            std::cout << "c: " << c << " v: " << disparity[image_cell] << std::endl;
			if(frame_number == 0) {

                bgValues[c].currentValue[0].d = (ushort)disparity[image_cell];
                //std::cout << "bgcurrent value: "<< bgValues[c].currentValue[0].d << " " << disparity[image_cell] << std::endl;
				bgValues[c].counter[0] = 1;

				for(int i = 1; i < numSamples; i++)	{
					bgValues[c].counter[i] = 0;
				}

			}//if(frame_number == 0)
			else {
				for(int i = 0; i <= frame_number; i++) {
                    currentPixel.d = (ushort)disparity[image_cell];
                   // std::cout << "current: "<< currentPixel.d << " " << disparity[image_cell] << std::endl;
					if(bgValues[c].counter[i] == 0)	{
						bgValues[c].currentValue[i].d = currentPixel.d;
						bgValues[c].counter[i]++;
						break;
					}
					else if(abs(currentPixel.d - bgValues[c].currentValue[i].d <= binWidth))
					{
						int den = (bgValues[c].counter[i] + 1);
						bgValues[c].currentValue[i].d =
							(bgValues[c].currentValue[i].d * bgValues[c].counter[i]	+ currentPixel.d) / den;
						bgValues[c].counter[i]++;
						break;
					}
				}//for(int i = 0; i < numSamples; i++)
				if(do_background) {
					int index = 0;
					int max_counter = -1;
					bool set_isValid = false;
					for(int i = 0; i < numSamples; i++){
						if(bgValues[c].counter[i] == 0) {
							bgModel[c].isValid[index] = false;
							set_isValid = true;
							break;
						}
						else if(index == maxBgValues) {
							break;
						}
						else if(bgValues[c].counter[i] >= minBinHeight) {
							if(bgValues[c].counter[i] > max_counter) {
								max_counter = bgValues[c].counter[i];
								bgModel[c].values[index].d = bgModel[c].values[0].d;
								bgModel[c].isValid[index] = true;
								bgModel[c].counter[index] = bgModel[c].counter[0];

								bgModel[c].values[0].d = bgValues[c].currentValue[i].d;

								bgModel[c].counter[0] = bgValues[c].counter[i];
							}
							else {
								bgModel[c].values[index].d = bgValues[c].currentValue[i].d;
								bgModel[c].isValid[index] = true;

								bgModel[c].counter[index] = bgValues[c].counter[i];
							}
							index++;
						}
					}
					if(!set_isValid) {
						int step = numSamples / maxBgValues;
						for(int n = 0, index = 0; n < maxBgValues; n++, index += step) {
							bgModel[c].values[n].d = bgValues[c].currentValue[index].d;
							bgModel[c].isValid[n] = true;

							bgModel[c].counter[n] = bgValues[c].counter[index];
						}
					}
					else if(index == 0) {
						int max_counter = -1;
						for(int i = 0; i < numSamples; i++){
							if(bgValues[c].counter[i] == 0) {
								bgModel[c].isValid[index] = false;
								break;
							}
							else if(index == maxBgValues) {
								break;
							}
							else if(bgValues[c].counter[i] >= (minBinHeight - 1)) {
								if(bgValues[c].counter[i] > max_counter) {
									max_counter = bgValues[c].counter[i];
									bgModel[c].values[index].d = bgModel[c].values[0].d;
									bgModel[c].isValid[index] = true;

									bgModel[c].counter[index] = bgModel[c].counter[0];
									bgModel[c].values[0].d = bgValues[c].currentValue[i].d;

									bgModel[c].counter[0] = bgValues[c].counter[i];
								}
								else {
									bgModel[c].values[index].d = bgValues[c].currentValue[i].d;
									bgModel[c].isValid[index] = true;

									bgModel[c].counter[index] = bgValues[c].counter[i];
								}
							}
							index++;
						}// for i < numSamples
					}

				}//if(frame_number == numSamples - 1)
			}//else --> if(frame_number == 0)
		}//for
	}//for


	if(frame_number == (numSamples - 1)) {

		bg_reset = false;
        int offset = 0;


        for(int i = 0; i < frameHeight; ++i){
            for(int j = 0; j < frameWidth; ++j) {
                    memcpy(&bgStereoImageMat.data[offset],&bgModel[i * frameWidth + j].values[0].d, sizeof(short));
                    offset += sizeof(short);
            }
        }

//        for(int c = 0; c < numPixels; c++)
//            bgStereoImageMat.data[offset] = (uchar)bgModel[c].values[0].d;
//            offset += sizeof(short);
////			bgStereoImage->imageData[c] = bgModel[c].values[0].d;
//		}

//        cv::imshow("ggg", bgStereoImageMat);
//        cv::waitKey(0);
        //bgStereoImageMat = cv::Mat(bgStereoImage);



        if(bgStereoFilename != "") {
            cout << "Saving Stereo Background Data in " <<
                    bgStereoFilename.c_str() << "...";
            cout.flush();
            ofstream file;
            file.open(bgStereoFilename.c_str());
            if(file.is_open()) {
                int c = 0;
                for(int i = 0; i<frameHeight; i++) {
                    for(int j = 0; j<frameWidth; j++, c++) {
                        for(int e = 0; e < maxBgValues; e++) {
                            if(!bgModel[c].isValid[e]) {
                                file<<endl;
                                break;
                            }
                            file<<(int)bgModel[c].values[e].d<<" ";
                            if(e == (maxBgValues - 1)) {
                                file<<endl;
                            }
                        }
                    }
                }
                file.close();
                cout << "done." << endl;
            }
            else {
                cout << endl;
                cout << "Unable to write on " << bgStereoFilename << endl;
                cout << "Aborting..." << endl;
                abort();
            }
            bgStereoFilename = "";
        }//if bgStereoFilename
    }
}

void IMBS::createBackground(int frame_number, int image_interval, bool do_background) {
	if(bgSample == NULL) {
		cerr << "createBackground -- an error occurred: " <<
				" unable to retrieve frame no. " << frame_number << endl;
		abort();
    }

	int h_interval = frameHeight/totalImageParts;
	int h = (image_interval * h_interval);
	int image_index = h * frame->widthStep;
	Pixel currentPixel;

	for(int t = 0; t < h_interval; h++, t++, image_index += frame->widthStep) {
		for(int j = 0, image_cell = image_index, c = image_index / 3;
				j < frameWidth;
				j++, image_cell += 3, c++)
		{
            if(frame_number == 0) {

               if((uchar)fgImage->imageData[c] == 255) {
            	   bgValues[c].currentValue[0].b = bgImage->imageData[image_cell];
				   bgValues[c].currentValue[0].g = bgImage->imageData[image_cell+1];
				   bgValues[c].currentValue[0].r = bgImage->imageData[image_cell+2];

               }
               else {
                    bgValues[c].currentValue[0].b =
                        (uchar)bgSample->imageData[image_cell];
                    bgValues[c].currentValue[0].g =
                        (uchar)bgSample->imageData[image_cell+1];
                    bgValues[c].currentValue[0].r =
                        (uchar)bgSample->imageData[image_cell+2];
               }
               bgValues[c].counter[0] = 1;

               for(int i = 1; i < numSamples; i++)	{
            	   bgValues[c].counter[i] = 0;
               }

#if ANDREA
               if((uchar)fgImage->imageData[c] == 255) {
            	   bgValues[c].isForeground[0] = true;
               }
               else {
            	   bgValues[c].isForeground[0] = false;
               }
#endif

			}//if(frame_number == 0)
            else {
				for(int i = 0; i <= frame_number; i++) {

					if((uchar)fgImage->imageData[c] == 255) {
						if(bgModel[c].counter[i] != 0) {
							currentPixel.b = bgModel[c].values[i].b;
							currentPixel.g = bgModel[c].values[i].g;
							currentPixel.r = bgModel[c].values[i].r;

						}
						else {
							break;
						}
					}
					else {

						currentPixel.b = (uchar)(bgSample->imageData)[image_cell];
						currentPixel.g = (uchar)(bgSample->imageData)[image_cell+1];
						currentPixel.r = (uchar)(bgSample->imageData)[image_cell+2];
					}

					if(bgValues[c].counter[i] == 0)	{

						bgValues[c].currentValue[i].r = currentPixel.r;
						bgValues[c].currentValue[i].g = currentPixel.g;
						bgValues[c].currentValue[i].b = currentPixel.b;
						bgValues[c].counter[i]++;

#if ANDREA
						if(fgImage->imageData[c] == (uchar)255) {
							bgValues[c].isForeground[i] = true;

						}
						else {
							bgValues[c].isForeground[i] = false;
						}
#endif
						break;
					}
					else if( abs(currentPixel.r - bgValues[c].currentValue[i].r) <= binWidth &&
						 abs(currentPixel.g - bgValues[c].currentValue[i].g) <= binWidth &&
						 abs(currentPixel.b - bgValues[c].currentValue[i].b) <= binWidth)
					{
						int den = (bgValues[c].counter[i] + 1);

						bgValues[c].currentValue[i].r =
							(bgValues[c].currentValue[i].r * bgValues[c].counter[i]	+ currentPixel.r) / den;
						bgValues[c].currentValue[i].g =
							(bgValues[c].currentValue[i].g * bgValues[c].counter[i]	+ currentPixel.g) / den;
						bgValues[c].currentValue[i].b =
							(bgValues[c].currentValue[i].b * bgValues[c].counter[i]	+ currentPixel.b) / den;
						bgValues[c].counter[i]++;

#if ANDREA
					   if((uchar)fgImage->imageData[c] == 255) {
						   bgValues[c].isForeground[i] = true;
					   }
#endif

					   break;
					}
				}//for(int i = 0; i < numSamples; i++)






				if(do_background) {
                    //prevBgValues[c].currentValue = bgValues[c].currentValue;
                    //prevBgValues[c].counter = bgValues[c].counter;
                    //prevBgValues[c].isForeground = bgValues[c].isForeground;

					int index = 0;
					int max_counter = -1;
					bool set_isValid = false;
					for(int i = 0; i < numSamples; i++){
						if(bgValues[c].counter[i] == 0) {
							bgModel[c].isValid[index] = false;
							set_isValid = true;
							break;
						}
						else if(index == maxBgValues) {
							break;
						}
                        else if(bgValues[c].counter[i] >= minBinHeight) {

#if ANDREA
                            if(fgImage->imageData[c] == 120) {
                                for(int n = 0; n < maxBgValues; n++) {
                                    if(!bgModel[c].isValid[n]) {
                                        break;
                                    }
                                    int d = max(abs(bgModel[c].values[n].b - bgValues[c].currentValue[i].b),
                                        abs(bgModel[c].values[n].g - bgValues[c].currentValue[i].g) );
                                    d = max(d, abs(bgModel[c].values[n].r - bgValues[c].currentValue[i].r) );
                                    if(d < fgThreshold){
                                        bgModel[c].isForeground[n] = false;
                                        bgValues[c].isForeground[i] = false;
                                    }
                                }
                            }
#endif

                            if(bgValues[c].counter[i] > max_counter) {
								max_counter = bgValues[c].counter[i];
								bgModel[c].values[index].r = bgModel[c].values[0].r;
								bgModel[c].values[index].g = bgModel[c].values[0].g;
								bgModel[c].values[index].b = bgModel[c].values[0].b;
								bgModel[c].isValid[index] = true;
#if ANDREA
								bgModel[c].isForeground[index] = bgModel[c].isForeground[0];
#endif
								bgModel[c].counter[index] = bgModel[c].counter[0];


								bgModel[c].values[0].r = bgValues[c].currentValue[i].r;
								bgModel[c].values[0].g = bgValues[c].currentValue[i].g;
								bgModel[c].values[0].b = bgValues[c].currentValue[i].b;

#if ANDREA

								bgModel[c].isForeground[0] = bgValues[c].isForeground[i];
#endif
								bgModel[c].counter[0] = bgValues[c].counter[i];

							}
                            else {
								bgModel[c].values[index].r = bgValues[c].currentValue[i].r;
								bgModel[c].values[index].g = bgValues[c].currentValue[i].g;
								bgModel[c].values[index].b = bgValues[c].currentValue[i].b;
								bgModel[c].isValid[index] = true;

							}
							index++;
						}
					}
                    if(!set_isValid) {

                        int step = numSamples / maxBgValues;
						for(int n = 0, index = 0; n < maxBgValues; n++, index += step) {

                                bgModel[c].values[n].r = bgValues[c].currentValue[index].r;
                                bgModel[c].values[n].g = bgValues[c].currentValue[index].g;
                                bgModel[c].values[n].b = bgValues[c].currentValue[index].b;

                                bgModel[c].isValid[n] = true;

                                bgModel[c].counter[n] = bgValues[c].counter[index];
						}
					}
                    else if(index == 0) {
						int max_counter = -1;
						for(int i = 0; i < numSamples; i++){
							if(bgValues[c].counter[i] == 0) {
								bgModel[c].isValid[index] = false;
								break;
							}
							else if(index == maxBgValues) {
								break;
							}
                            else if(bgValues[c].counter[i] >= (minBinHeight - 1)) {
                                if(bgValues[c].counter[i] > max_counter ) {
									max_counter = bgValues[c].counter[i];

									bgModel[c].values[index].r = bgModel[c].values[0].r;
									bgModel[c].values[index].g = bgModel[c].values[0].g;
									bgModel[c].values[index].b = bgModel[c].values[0].b;
									bgModel[c].isValid[index] = true;

									bgModel[c].counter[index] = bgModel[c].counter[0];

									bgModel[c].values[0].r = bgValues[c].currentValue[i].r;
									bgModel[c].values[0].g = bgValues[c].currentValue[i].g;
									bgModel[c].values[0].b = bgValues[c].currentValue[i].b;

									bgModel[c].counter[0] = bgValues[c].counter[i];
								}
                                else {
									bgModel[c].values[index].r = bgValues[c].currentValue[i].r;
									bgModel[c].values[index].g = bgValues[c].currentValue[i].g;
									bgModel[c].values[index].b = bgValues[c].currentValue[i].b;
									bgModel[c].isValid[index] = true;

									bgModel[c].counter[index] = bgValues[c].counter[i];
								}
							}
							index++;
						}// for i < numSamples
					}

				}//if(frame_number == numSamples - 1)
			}//else --> if(frame_number == 0)
		}//for
	}//for

	if(frame_number == (numSamples - 1)) {

		bg_reset = false;

		for(int c = 0, t = 0; c < numPixels; c++, t+=3) {
                bgImage->imageData[t] = bgModel[c].values[0].b;
                bgImage->imageData[t+1] = bgModel[c].values[0].g;
                bgImage->imageData[t+2] = bgModel[c].values[0].r;
		}

//        bgModelPrev->values = bgModel->values;
//        bgModelPrev->isValid = bgModel->isValid;
//        bgModelPrev->counter = bgModel->counter;
//        bgModelPrev->isForeground = bgModelPrev->isForeground;

//        doPrevBgModel = true;

		if(bgFilename != NULL) {
			ofstream file;
			file.open(bgFilename->c_str());
			int c = 0;
			for(int i = 0; i<frameHeight; i++) {
				for(int j = 0; j<frameWidth; j++, c++) {
					for(int e = 0; e < maxBgValues; e++) {
						if(!bgModel[c].isValid[e]) {
							file<<endl;
							break;
						}
						file<<(int)bgModel[c].values[e].r<<" ";
						file<<(int)bgModel[c].values[e].g<<" ";
						file<<(int)bgModel[c].values[e].b<<" ";
						if(e == (maxBgValues - 1)) {
							file<<endl;
						}
					}
				}
			}
			file.close();
			bgFilename = NULL;
		}//if bgFilename
	}
}

bool onlyone = false;


void calcProjection(cv::Point2f &point, const cv::Mat &H, double &x, double &y) {

    x =	(point.x*H.at<double>(0,0)+point.y*H.at<double>(0,1)+H.at<double>(0,2))/
            (point.x*H.at<double>(2,0)+point.y*H.at<double>(2,1)+H.at<double>(2,2));
    y =	(point.x*H.at<double>(1,0)+point.y*H.at<double>(1,1)+H.at<double>(1,2))/
            (point.x*H.at<double>(2,0)+point.y*H.at<double>(2,1)+H.at<double>(2,2));
}

void IMBS::process() {
	updateBackground();

	if(bgModel[0].isValid[0]) {

		getForeground();

        hsvSuppression(); //lavora su filtered FG

		cvMorphologyEx(fgFilteredImage, fgFilteredImage, NULL, NULL, CV_MOP_OPEN, 1);

		cvMorphologyEx(fgFilteredImage, fgFilteredImage, NULL, NULL, CV_MOP_CLOSE, 1);

		areaThresholding(area);

        depthFilter();//remove objects far from the camera (4 meters)

        heightFilter();//remove objects over 70 cm height


#if SAVE_DATA

        if(grabber->frameNumber == 2074) {
            CBlobResult blobs = CBlobResult( fgFilteredImage, NULL, 0 );
            cv::Mat tempMat = cv::Mat::zeros(cv::Size(640, 480), CV_8UC1);
            CBlob currentBlob;
            int numBlobs = blobs.GetNumBlobs();
            std::ofstream ff("frame1943.txt", std::ios::out);

            for(int n = 0; n < numBlobs; n++) {
                currentBlob = blobs.GetBlob(n);

                int max_x = (int)currentBlob.MaxX();
                int min_x = (int)currentBlob.MinX();
                int max_y = (int)currentBlob.MaxY();
                int min_y = (int)currentBlob.MinY();
                for(int i = min_y; i < max_y; i++) {
                    for(int j = min_x; j < max_x; j++) {
                        if((uchar)fgFilteredImage->imageData[j+frameWidth * ( i % frameHeight)] == 255) {
                            if(i >= static_cast<imagestereograbber *>(grabber)->ly &&
                                    i<= static_cast<imagestereograbber *>(grabber)->ry &&
                                    j >= static_cast<imagestereograbber *>(grabber)->lx &&
                                    j <= static_cast<imagestereograbber *>(grabber)->rx) {

                                    tempMat.at<uchar>(i,j) = 255;
                                    double x, y, z;
                                    //double theta = -28.5;
                                    //double t = 520;
                                    static_cast<imagestereograbber*>(grabber)->get3DPoint(j, i, &x, &z, &y);
                                    std::cout << x << " " << z << " " << y << std::endl;
                                    if(x == -0) x = 0;
                                    double x1, y1;
                                    cv::Point2f p(x*0.1,z*0.1);
                                    calcProjection(p, static_cast<imagestereograbber*>(grabber)->homography, x1, y1);
                                    ff << j << " " << i << " " << x1*0.01 << " " << y1*0.01 << " " << 0 << std::endl;
                                    //ff << x << " " << y << " " << z << std::endl;

                             }
                        }
                    }
                }
            }

            //cv::imshow("temp", tempMat);
              ff.close();
              cvSaveImage("fg.bmp", fgFilteredImage);

              exit(-1);
        }
#endif


	}

}

void IMBS::heightFilter() {

	cv::Mat depthMap = static_cast<imagestereograbber*>(grabber)->getDepth16bit();

    for(int i = 0; i < frameHeight; ++i) {
        for(int j = 0; j < frameWidth; ++j) {
            if((uchar)fgFilteredImage->imageData[j+frameWidth * (i % frameHeight)] == 255
                    && depthMap.at<ushort>(i,j) > 30 && depthMap.at<ushort>(i,j) < 3500) {
                fgFilteredImage->imageData[j+frameWidth * (i % frameHeight)] = 255;
            } else {
                fgFilteredImage->imageData[j+frameWidth * (i % frameHeight)] = 0;
            }
        }
    }


    cv::Mat temp = cv::Mat::zeros(cv::Size(640, 480), CV_8UC1);


    CBlobResult blobs = CBlobResult( fgFilteredImage, NULL, 0 );
    CBlob currentBlob;
    int numBlobs = blobs.GetNumBlobs();


    cvZero(depthBlobs);

    for(int k = 0; k < numBlobs; k++) {
        currentBlob = blobs.GetBlob(k);

        int max_x = (int)currentBlob.MaxX();
        int min_x = (int)currentBlob.MinX();
        int max_y = (int)currentBlob.MaxY();
        int min_y = (int)currentBlob.MinY();


        int width = max_x - min_x;
        int height = max_y - min_y;

        int area = width * height;

        if(height > 0 && width > 0 &&  area > 3000) {

        	int cnt_i = 0;
        	int cnt_j = 0;
        	int sumY = 0;

            for(int i = min_y; i < max_y; ++i) {

            	if(cnt_i > 10)
            		break;

                for(int j = min_x; j < max_x; ++j) {
                    if((uchar)fgFilteredImage->imageData[j+frameWidth * (i % frameHeight)] == 255) {
                        double X, Y, Z;
                        static_cast<imagestereograbber*>(grabber)->get3DPoint(j,i,&X,&Z,&Y);
                        if(Z > 500 && Z < 3500) {
                            temp.at<uchar>(i,j) = 255;
                    		sumY += Y;
                    		cnt_j++;
                    	}

                    }
                }
                cnt_i++;
            }

            if(cnt_j == 0) {
            	cvRectangle(fgFilteredImage, cvPoint(min_x,min_y), cvPoint(max_x,max_y), CV_RGB(0,0,0), -1);
            }
            else {
            	double altezza = sumY / cnt_j;
                std::cout << "frame: " << grabber->frameNumber << " altezza: " << altezza << std::endl;
            	if(altezza > 600) {
            		cvRectangle(fgFilteredImage, cvPoint(min_x,min_y), cvPoint(max_x,max_y), CV_RGB(0,0,0), -1);
            	}
            }

            for(int i = min_y; i < max_y; ++i) {
				for(int j = min_x; j < max_x; ++j) {
					if((uchar)fgFilteredImage->imageData[j+frameWidth * (i % frameHeight)] == 255) {
                        double X, Y, Z;
                        static_cast<imagestereograbber*>(grabber)->get3DPoint(j,i,&X,&Z,&Y);
                        if(Z > 500 && Z < 3500) {
							if(Y  < 30) {
								fgFilteredImage->imageData[j+frameWidth * (i % frameHeight)] = 0;
							}
						}

					}
				}

			}


        } else {
            cvRectangle(fgFilteredImage, cvPoint(min_x,min_y), cvPoint(max_x,max_y), CV_RGB(0,0,0), -1);
        }
    }




    findBall ball(grabber->frame, fgFilteredImage);
    ball.process();

    cvMorphologyEx(fgFilteredImage, fgFilteredImage, NULL, NULL, CV_MOP_OPEN, 1);

    cvMorphologyEx(fgFilteredImage, fgFilteredImage, NULL, NULL, CV_MOP_CLOSE, 1);


    blobs = CBlobResult( fgFilteredImage, NULL, 0 );
    numBlobs = blobs.GetNumBlobs();

    //std::cout << "*******" << std::endl;
    for(int k = 0; k < numBlobs; k++) {
        currentBlob = blobs.GetBlob(k);

        int max_x = (int)currentBlob.MaxX();
        int min_x = (int)currentBlob.MinX();
        int max_y = (int)currentBlob.MaxY();
        int min_y = (int)currentBlob.MinY();


        int width = max_x - min_x;
        int height = max_y - min_y;

        if(height > 0 && width > 0) {
            int count = 0;
            double x = .0, y = .0;

            for(int i = min_y; i < max_y; ++i) {

                for(int j = min_x; j < max_x; ++j) {
                    if((uchar)fgFilteredImage->imageData[j+frameWidth * (i % frameHeight)] == 255) {
                        x+=j;
                        y+=i;
                        count++;
                    }
                }
            }

            double X, Y, Z;
            static_cast<imagestereograbber*>(grabber)->get3DPoint(x/count, y/count,&X,&Z,&Y);
            double x1, y1;
            cv::Point2f p(x/count, y/count);
            calcProjection(p, static_cast<imagestereograbber*>(grabber)->homography, x1, y1);
            //std::cout << x1 << " " << y1 << " " << X << " " << Y << " " << Z << std::endl;
        }
    }
    //std::cout << "*******" << std::endl;

    /*if(pointBlobs.size() > 0) {


        cv::Mat HL = cv::Mat::zeros(cv::Size(3, 3), CV_64F);

        HL.at<double>(0,0) = 0.404156;
        HL.at<double>(0,1) = -4.37861;
        HL.at<double>(0,2) = 2200.43;
        HL.at<double>(1,0) = 5.05127;
        HL.at<double>(1,1) = 15.1579;
        HL.at<double>(1,2) = -894.026 ;
        HL.at<double>(2,0) = 0.00105902;
        HL.at<double>(2,1) = 0.0287815;
        HL.at<double>(2,2) = 1;

        cv::Mat field = cv::imread("field_1.png");

        std::vector<cv::Point2f>::iterator it;
        std::vector<CvScalar>::iterator its;

        for(it = pointBlobs.begin(), its = pointBlobsColor.begin(); it != pointBlobs.end(); it++, its++) {
            double x, y;
            calcProjection(*it, HL, x, y);
            CvScalar temp = *its;
            cv::circle(field, cv::Point2f(x, y), 4, cv::Scalar(temp.val[0], temp.val[1], temp.val[2]), -1);
        }

        cv::imshow("campo", field);
    }*/
}


void IMBS::depthFilter() {

	cv::Mat depthMat = static_cast<imagestereograbber*>(grabber)->getDepth16bit();
    cv::Mat newDepthMat = cv::Mat::zeros(depthMat.size(), CV_16UC1);

	cv::Mat coloredDepth = cv::Mat::zeros(depthMat.size(), CV_8UC3);
	for(int i = 0; i < depthMat.rows; ++i) {
		for(int j = 0; j < depthMat.cols; ++j) {
			ushort v = depthMat.at<ushort>(i,j);
            if(v > 4000) {
                coloredDepth.at<cv::Vec3b>(i,j)[2] = 0;
                coloredDepth.at<cv::Vec3b>(i,j)[1] = 0;
                coloredDepth.at<cv::Vec3b>(i,j)[0] = 0;
            }
            else if(v > 3500) {
                coloredDepth.at<cv::Vec3b>(i,j)[2] = 0;
                coloredDepth.at<cv::Vec3b>(i,j)[1] = 0;
                coloredDepth.at<cv::Vec3b>(i,j)[0] = 0;
			}
            else if(v > 3000) {
				coloredDepth.at<cv::Vec3b>(i,j)[2] = 255;
				coloredDepth.at<cv::Vec3b>(i,j)[1] = 255;
				coloredDepth.at<cv::Vec3b>(i,j)[0] = 255;
                newDepthMat.at<ushort>(i,j) = v;
			}
			else if(v > 2500) {
				coloredDepth.at<cv::Vec3b>(i,j)[2] = 0;
				coloredDepth.at<cv::Vec3b>(i,j)[1] = 0;
				coloredDepth.at<cv::Vec3b>(i,j)[0] = 255;
                newDepthMat.at<ushort>(i,j) = v;
			}
			else if(v > 2000) {
				coloredDepth.at<cv::Vec3b>(i,j)[2] = 0;
				coloredDepth.at<cv::Vec3b>(i,j)[1] = 255;
				coloredDepth.at<cv::Vec3b>(i,j)[0] = 0;
                newDepthMat.at<ushort>(i,j) = v;
			}
			else if(v > 1500) {
				coloredDepth.at<cv::Vec3b>(i,j)[2] = 255;
				coloredDepth.at<cv::Vec3b>(i,j)[1] = 0;
				coloredDepth.at<cv::Vec3b>(i,j)[0] = 0;
                newDepthMat.at<ushort>(i,j) = v;
			}
			else if(v > 1000) {
				coloredDepth.at<cv::Vec3b>(i,j)[2] = 255;
				coloredDepth.at<cv::Vec3b>(i,j)[1] = 0;
				coloredDepth.at<cv::Vec3b>(i,j)[0] = 255;
                newDepthMat.at<ushort>(i,j) = v;
			}
			else if(v > 500) {
				coloredDepth.at<cv::Vec3b>(i,j)[2] = 0;
				coloredDepth.at<cv::Vec3b>(i,j)[1] = 255;
				coloredDepth.at<cv::Vec3b>(i,j)[0] = 255;
                newDepthMat.at<ushort>(i,j) = v;
			}
            else {
				coloredDepth.at<cv::Vec3b>(i,j)[2] = 0;
				coloredDepth.at<cv::Vec3b>(i,j)[1] = 0;
				coloredDepth.at<cv::Vec3b>(i,j)[0] = 0;
            }
		}
	}

	//cv::imshow("COLORED", coloredDepth);
    //cv::imshow("newDepth", newDepthMat);



    currentDisparity = new short[frameWidth * frameHeight * 2];
    for(int i = 0; i < frameHeight; ++i) {
        for(int j = 0; j < frameWidth; ++j) {
                currentDisparity[i * frameWidth + j] =  newDepthMat.at<ushort>(i,j);
        }
    }


    //CALCOLO IL FOREGROUND STEREO
    fgStereo = new short[numPixels*2];
    getStereoForeground();

    CBlobResult blobs = CBlobResult( fgFilteredImage, NULL, 0 );
    CBlob currentBlob;
    int numBlobs = blobs.GetNumBlobs();


    cvZero(fgColoredImage);
    for(int n = 0; n < numBlobs; n++) {
        currentBlob = blobs.GetBlob(n);

        int max_x = (int)currentBlob.MaxX();
        int min_x = (int)currentBlob.MinX();
        int max_y = (int)currentBlob.MaxY();
        int min_y = (int)currentBlob.MinY();
        double h = .0;
        for(int i = min_y; i < max_y; ++i) {
            for(int j = min_x; j < max_x; ++j) {
                if((uchar)fgFilteredImage->imageData[j+frameWidth * ( i % frameHeight)] == 255) {
                	//align RGB image and depth map
                	if(i >= static_cast<imagestereograbber *>(grabber)->ly &&
                            i<= static_cast<imagestereograbber *>(grabber)->ry &&
                            j >= static_cast<imagestereograbber *>(grabber)->lx &&
                            j <= static_cast<imagestereograbber *>(grabber)->rx) {


                		cvSet2D(fgColoredImage, i, j, coloredDepth.at<cv::Vec3b>(i,j));

#if TOGLI
                        if(fgStereo[j+frameWidth * ( i % frameHeight)] != 0) {
                            double x, y, z;
                            double theta = -30;
                            double t = 330;
                            static_cast<imagestereograbber*>(grabber)->get3DPoint(j, i, &x, &y, &z, theta, t);
                            if(z > h) h = z;
                        }
#endif
                    }
                }

            }
        }


#if TOGLI
        if(h > 700 || h <= 0) {
            cvRectangle(fgFilteredImage, cvPoint(min_x, min_y),	cvPoint(max_x, max_y), cvScalar(0), -1);
        }
#endif

    }
    delete fgStereo;
}


void IMBS::processStereo() {
	updateBackground();
}


bool IMBS::checkBlobArea(CBlobResult playerBlobs, double area) {
	int n = playerBlobs.GetNumBlobs();
	CBlob currentBlob;
	CBlobGetArea getArea;
	for(int i = 0; i < n; ++i ) {
		currentBlob = playerBlobs.GetBlob(i);
		if((double)getArea(currentBlob) == area) {
			return true;
		}
	}
	return false;
}

void IMBS::hsvSuppression() {

	cvCopy(fgImage, fgFilteredImage);

	IplImage* frame_32f = cvCreateImage(cvSize(frame->width, frame->height), IPL_DEPTH_32F, 3);
	cvConvertScale(frame, frame_32f, 0.003921568627451);

	cvCvtColor(frame_32f, hsvFrame, CV_BGR2HSV);

	float h_i, s_i, v_i;
	float h_b, s_b, v_b;
	float h_diff, s_diff, v_ratio;

	for(int counter = 0, counter_hsv = 0; counter < numPixels; counter++, counter_hsv+=3) {
		if(fgImage->imageData[counter]) {
			for(int n = 0; n < maxBgValues; n++) {
				if(!bgModel[counter].isValid[n]) {
					break;
				}

				float f_b = ((uchar)bgModel[counter].values[n].b)/255.f;
				float f_g = ((uchar)bgModel[counter].values[n].g)/255.f;
				float f_r = ((uchar)bgModel[counter].values[n].r)/255.f;

				float * data = (float *)rgbMat->imageData;
				data[0] = f_b;
				data[1] = f_g;
				data[2] = f_r;

				cvCvtColor(rgbMat, hsvMat, CV_BGR2HSV);

				data = (float *)hsvMat->imageData;
				h_b = data[0];
				s_b = data[1];
				v_b = data[2];

				data = (float *)hsvFrame->imageData;
				h_i = data[counter_hsv];
				s_i = data[counter_hsv + 1];
				v_i = data[counter_hsv + 2];

				v_ratio = v_i / v_b;
				s_diff = abs(s_i - s_b);
				h_diff = min( abs(h_i - h_b), 360 - abs(h_i - h_b));

				if( v_ratio >= alpha && v_ratio < beta && s_diff <= tau_s && h_diff <= tau_h ) {
					fgFilteredImage->imageData[counter] = 0;
					break;
				}

				//else if( v_ratio >= beta && v_ratio <= 1.15 && s_diff < 0.1 && h_diff <= tau_h) {
					//fgImage->imageData[counter] = 70;
					//break;
				//}

			}
		}
	}
	cvReleaseImage(&frame_32f);
}

bool IMBS::loadBackground(const char* filename) {
	string line;
	ifstream file(filename);
	int c = 0;
	if(file.is_open()) {
		while(!file.eof()) {
			getline(file, line);

			int n = 0;

			while(line.length() > 1){

				int index = line.find_first_of(" ");
				string red = line.substr(0, index);
				int r;
				istringstream ss_r(red);
				ss_r >> r;
				line.erase(0, index+1);

				index = line.find_first_of(" ");
				string green = line.substr(0, index);
				int g;
				istringstream ss_g(green);
				ss_g >> g;
				line.erase(0, index+1);

				index = line.find_first_of(" ");
				string blue = line.substr(0, index);
				int b;
				istringstream ss_b(blue);
				ss_b >> b;
				line.erase(0, index+1);

				bgModel[c].values[n].r = r;
				bgModel[c].values[n].g = g;
				bgModel[c].values[n].b = b;
				bgModel[c].isValid[n] = true;
				bgModel[c].counter[n] = minBinHeight;

				if(n == 0) {
					bgImage->imageData[c*3] = (uchar)bgModel[c].values[0].b;
					bgImage->imageData[c*3+1] = (uchar)bgModel[c].values[0].g;
					bgImage->imageData[c*3+2] = (uchar)bgModel[c].values[0].r;
				}
				n++;
			}
			c++;
	    }
	    file.close();
	    return true;
	}
	else {
		return false;
	}
}

bool IMBS::loadStereoBackground(const char* filename) {
	string line;
	ifstream file(filename);
	int c = 0;
	if(file.is_open()) {
		while(!file.eof()) {
			getline(file, line);

			int n = 0;

			while(line.length() > 1){

				int index = line.find_first_of(" ");
				string disp = line.substr(0, index);
				int d;
				istringstream ss_d(disp);
				ss_d >> d;
				line.erase(0, index+1);

				bgModel[c].values[n].d = d;
				bgModel[c].isValid[n] = true;
				bgModel[c].counter[n] = minBinHeight;

				if(n == 0) {
					bgStereoImage->imageData[c] = (uchar)bgModel[c].values[0].d;
				}
				n++;
			}
			c++;
	    }
	    file.close();
	    return true;
	}
	else {
		return false;
	}
}

void IMBS::areaThresholding(double area) {
	double area_bg_reset = 0.6 * numPixels;
	double totalArea = 0;
    CBlobResult blobs = CBlobResult( fgFilteredImage, NULL, 0 );
    CBlob currentBlob;
    int numBlobs = blobs.GetNumBlobs();
    int step_fg = fgFilteredImage->widthStep/sizeof(uchar);

    for(int n = 0; n < numBlobs; n++) {
        currentBlob = blobs.GetBlob(n);

        currentBlob.FillBlob( fgFilteredImage, CV_RGB(0,0,0));

        int max_x = (int)currentBlob.MaxX();
        int min_x = (int)currentBlob.MinX();
        int max_y = (int)currentBlob.MaxY();
        int min_y = (int)currentBlob.MinY();

		CBlobGetArea getArea;
		double blobArea = getArea(currentBlob);
        if(area > blobArea) {
			cvRectangle(fgFilteredImage, cvPoint(min_x, min_y),	cvPoint(max_x, max_y), cvScalar(0), -1);
		}
	}
}

